function love.run()
  _r = math.random
  _f = math.floor
  _c = math.cos
  _s = math.sin
  _sq = math.sqrt
  _at2 = math.atan2
  _d2r = math.rad
  _r2d = math.deg
  _m = math.mod
  pi = math.pi
  gr = love.graphics
  ti = love.timer
  ms = love.mouse
  kb = love.keyboard
  math.randomseed(os.time())

	love.load(arg)
	while true do
	  fps = ti.getFPS()
	  dt = ti.getDelta()
		ti.step()
		time = ti.getTime()
		gr.clear()
		love.update(dt)
		love.draw(dt)
		if love.event then
			for e,a,b,c in love.event.poll() do
				if e == "q" then
					if love.audio then love.audio.stop() end
				return
				end
				love.handlers[e](a,b,c)
			end
		end
		ti.sleep(1)
		gr.present()
	end
end

function love.load()
  gr.setLineWidth(1)

	imgMap = gr.newImage("map.gif")

	imgGhost = gr.newImage("ghost.gif")
	qGhost = {
	  [1] = gr.newQuad(0,0,16,16,32,32),
	  [2] = gr.newQuad(0,16,16,16,32,32)
	}
	qEyes = {
	  [1] = gr.newQuad(16,0,16,8,32,32),
	  [2] = gr.newQuad(16,8,16,8,32,32),
	  [3] = gr.newQuad(16,16,16,8,32,32),
	  [4] = gr.newQuad(16,24,16,8,32,32)
	}

	imgPac = gr.newImage("pacman.gif")
	qPac = {
	  [1] = gr.newQuad(0,0,15,15,64,16),
	  [2] = gr.newQuad(16,0,15,15,64,16),
	  [3] = gr.newQuad(32,0,15,15,64,16),
	}

	font = gr.newFont(love._vera_ttf, 10)

  empty = 0
  wall = 1
  pellet = 2
  power = 3

  stationary = 0
  left = 1
  up = 2
  right = 3
  down = 4

  chase = 1
  scatter = 2
  frightened = 3

  lv = { map = {}, mode = chase, pellets = 0, frightcounter = 0 }
  lv.map.tile = {}
  for x=-3,28+2 do
    lv.map.tile[x] = {}
    for y=-3,36+2 do
      lv.map.tile[x][y] = wall
    end
  end
  lv.map.tile[-1][17] = empty
  lv.map.tile[28][17] = empty

	local mFile = "lv.map"
	if love.filesystem.exists(mFile) then
		local i = 0
		local mD = {}
		for line in love.filesystem.lines(mFile) do
			mD[i] = line
			i = i + 1
			--print(line)
		end

    for y=0,36-1 do
  		for x=0,28-1 do
				lv.map.tile[x][y] = tonumber(string.sub(mD[y], x+1, x+1))
			end
		end
	else
		error("Level Map NOT FOUND!! DUN DUN DUN!!!")
	end

  player = {
    x = 13,
    y = 26,
    ox = 4,
    oy = 0,
    mx = 13,
    my = 26,
    lives = 3,
    started = false,
    tomove = 0,
    facing = left,
    speed = 48,
    enteredTile = function()
      if player.x < 0 and player.y == 17 then
        player.x = 28
      elseif player.x > 27 and player.y == 17 then
        player.x = -1
      end
      if lv.map.tile[player.x][player.y] == pellet then
        lv.map.tile[player.x][player.y] = 0
        player.eatPellet()
      elseif lv.map.tile[player.x][player.y] == power then
        lv.map.tile[player.x][player.y] = 0
        player.eatPellet(true)
      end

      if player.tomove == left then
        if lv.map.tile[player.x-1][player.y] ~= 1 then
          player.ox = 8
          player.tomove = left
        else
          player.tomove = stationary
        end
      elseif player.tomove == right then
        if lv.map.tile[player.x+1][player.y] ~= 1 then
          player.ox = -8
          player.tomove = right
        else
          player.tomove = stationary
        end
      elseif player.tomove == up then
        if lv.map.tile[player.x][player.y-1] ~= 1 then
          player.oy = 8
          player.tomove = up
        else
          player.tomove = stationary
        end
      elseif player.tomove == down then
        if lv.map.tile[player.x][player.y+1] ~= 1 then
          player.oy = -8
          player.tomove = down
        else
          player.tomove = stationary
        end
      end
    end,
    leftTile = function()
      if player.tomove == left then
        if lv.map.tile[player.x-1][player.y] ~= 1 then
          player.x = player.x - 1
        else
          player.ox = 0
        end
      elseif player.tomove == right then
        if lv.map.tile[player.x+1][player.y] ~= 1 then
          player.x = player.x + 1
        else
          player.ox = 0
        end
      elseif player.tomove == up then
        if lv.map.tile[player.x][player.y-1] ~= 1 then
          player.y = player.y - 1
        else
          player.oy = 0
        end
      elseif player.tomove == down then
        if lv.map.tile[player.x][player.y+1] ~= 1 then
          player.y = player.y + 1
        else
          player.oy = 0
        end
      end
    end,
    eatPellet = function(pp)
      if pp then
        print(pp)
        lv.mode = frightened
        lv.frightcounter = 5
        for i, g in pairs(ghost) do
          if g.dir == left then
            g.dir = right
          elseif g.dir == right then
            g.dir = left
          elseif g.dir == up then
            g.dir = down
          elseif g.dir == down then
            g.dir = up
          end
        end
      end
    end
  }


  ghost = {}
  for g=1,4 do
    ghost[g] = {
      x = 0, y = 0,
      ox = 0, oy = 0,
      tx = 0, ty = 0,
      dir = 0,
      eaten = false,
      idling = true,
      inhouse = true,
      speed = 64,
      c = {244,224,222},
      go = function(dir)
        if dir == up then
          g.oy = 8
          g.y = g.y - 1
          g.dir = up
        elseif dir == left then
          g.ox = 8
          g.x = g.x - 1
          g.dir = left
        elseif dir == down then
          g.oy = -8
          g.y = g.y + 1
          g.dir = down
        elseif dir == right then
          g.ox = -8
          g.x = g.x + 1
          g.dir = right
        end
      end,
      reachedTile = function(n)
        g = ghost[n]
        if g.x < 0 and g.y == 17 then
          g.x = 28
        elseif g.x > 27 and g.y == 17 then
          g.x = -1
        end
        if g.eaten then
          g.tx = 13
          g.ty = 14
        elseif lv.mode == chase then
          if n == 1 then
            g.tx = player.x
            g.ty = player.y
          elseif n == 2 then
            if player.tomove == left then
              g.tx = player.x-4
              g.ty = player.y
            elseif player.tomove == right then
              g.tx = player.x+4
              g.ty = player.y
            elseif player.tomove == up then
              g.tx = player.x-4
              g.ty = player.y-4
            elseif player.tomove == down then
              g.tx = player.x
              g.ty = player.y+4
            end
          elseif n == 3 then
            local nx, ny = 0, 0
            if player.tomove == left then
              nx = player.x-2
              ny = player.y
            elseif player.tomove == right then
              nx = player.x+2
              ny = player.y
            elseif player.tomove == up then
              nx = player.x-2
              ny = player.y-2
            elseif player.tomove == down then
              nx = player.x
              ny = player.y+2
            end
            mx = (ghost[1].x - nx) * 2
            my = (ghost[1].y - ny) * 2
            g.tx = mx
            g.ty = my
          elseif n == 4 then
            if distanceFrom(player.x,player.y,g.x,g.y) <= 8 then
              g.tx = 2
              g.ty = 35
            else
              g.tx = player.x
              g.ty = player.y
            end
          end
          if g.tx < 0 then g.tx = 0 end
          if g.tx > 27 then g.tx = 27 end
          if g.ty < 0 then g.ty = 0 end
          if g.ty > 35 then g.ty = 35 end
        elseif lv.mode == scatter then
          if n == 1 then
            if level.pellets <= 20 then
              g.tx = player.x
              g.ty = player.y
            else
              g.tx = 25
              g.ty = 0
            end
          elseif n == 2 then
            g.tx = 2
            g.ty = 0
          elseif n == 3 then
            g.tx = 25
            g.ty = 35
          elseif n == 4 then
            g.tx = 2
            g.ty = 35
          end
        elseif lv.mode == frightened then
          g.tx = _r(0,23)
          g.ty = _r(0,35)
        end

        local tl, tr, tu, td = lv.map.tile[g.x-1][g.y], lv.map.tile[g.x+1][g.y], lv.map.tile[g.x][g.y-1], lv.map.tile[g.x][g.y+1]

        local dfl, dfr, dfu, dfd, fz
        dfl = distanceFrom((g.x-1)*8, g.y*8, g.tx*8, g.ty*8)
        dfr = distanceFrom((g.x+1)*8, g.y*8, g.tx*8, g.ty*8)
        dfu = distanceFrom(g.x*8, (g.y-1)*8, g.tx*8, g.ty*8)
        dfd = distanceFrom(g.x*8, (g.y+1)*8, g.tx*8, g.ty*8)
        fz = false
        if (g.x == 12 and g.y == 14) or (g.x == 15 and g.y == 14) or (g.x == 12 and g.y == 26) or (g.x == 15 and g.y == 26) then
          fz = true
        end

        --print("L "..tl, "R "..tr, "U "..tu, "D "..td, g.dir)
        if g.dir == left then

          if tl ~= 1 and tu == 1 and td == 1 then --L
            g.go(left)
          elseif tl == 1 and td == 1 and tu ~= 1 then --U
            g.go(up)
          elseif tl == 1 and td ~= 1 and tu == 1 then --D
            g.go(down)
          elseif tl == 1 and tu ~= 1 and td ~= 1 then --UD
            if dfl < dfd then
              g.go(up)
            else
              g.go(down)
            end
          elseif tl ~= 1 and tu ~= 1 and td == 1 then --LU
            if dfu < dfl and fz == false then
              g.go(up)
            else
              g.go(left)
            end
          elseif tl ~= 1 and td ~= 1 and tu == 1 then --LD
            if dfl < dfd then
              g.go(left)
            else
              g.go(down)
            end
          elseif tl ~= 1 and tu ~= 1 and td ~= 1 then --LUD
            if dfu < dfl and dfu < dfd and fz == false then
              g.go(up)
            elseif dfl < dfu and dfl < dfd then
              g.go(left)
            else
              g.go(down)
            end
          end

        elseif g.dir == right then

          if tr ~= 1 and tu == 1 and td == 1 then --R
            g.go(right)
          elseif tr == 1 and tu ~= 1 and td == 1 then --U
            g.go(up)
          elseif tr == 1 and tu == 1 and td ~= 1 then --D
            g.go(down)
          elseif tr == 1 and tu ~= 1 and td ~= 1 then --UD
            if dfu < dfd and fz == false then
              g.go(up)
            else
              g.go(down)
            end
          elseif tr ~= 1 and td == 1 and tu ~= 1 then --RU
            if dfu < dfr and fz == false then
              g.go(up)
            else
              g.go(right)
            end
          elseif tr ~= 1 and td ~= 1 and tu == 1 then --RD
            if dfd < dfr then
              g.go(down)
            else
              g.go(right)
            end
          elseif tr ~= 1 and tu ~= 1 and td ~= 1 then --RUD
            if dfu < dfr and dfu < dfd and fz == false then
              g.go(up)
            elseif dfd < dfu and dfd < dfr then
              g.go(down)
            else
              g.go(right)
            end
          end

        elseif g.dir == up then

          if tl == 1 and tr == 1 and tu ~= 1 then --U
            g.go(up)
          elseif tl ~= 1 and tr == 1 and tu == 1 then --L
            g.go(left)
          elseif tl == 1 and tr ~= 1 and tu == 1 then --R
            g.go(right)
          elseif tl ~= 1 and tr ~= 1 and tu == 1 then --LR
            if dfl < dfr then
              g.go(left)
            else
              g.go(right)
            end
          elseif tr == 1 and tl ~= 1 and tu ~= 1 then --LU
            if dfu < dfl and fz == false then
              g.go(up)
            else
              g.go(left)
            end
          elseif tr ~= 1 and tl == 1 and tu ~= 1 then --RU
            if dfu < dfr then
              g.go(up)
            else
              g.go(right)
            end
          elseif tr ~= 1 and tu ~= 1 and tl ~= 1 then --LRU
            if dfu < dfr and dfu < dfl and fz == false then
              g.go(up)
            elseif dfl < dfr and dfl < dfu then
              g.go(left)
            else
              g.go(right)
            end
          end

        elseif g.dir == down then

          if tl == 1 and tr == 1 and td ~= 1 then --D
            g.go(down)
          elseif tr == 1 and td == 1 and tl ~= 1 then --L
            g.go(left)
          elseif tl == 1 and td == 1 and tr ~= 1 then --R
            g.go(right)
          elseif tl ~= 1 and td == 1 and tr ~= 1 then --LR
            if dfl < dfr then
              g.go(left)
            else
              g.go(right)
            end
          elseif tl ~= 1 and tr == 1 and td ~= 1 then --LD
            if dfl < dfd then
              g.go(left)
            else
              g.go(down)
            end
          elseif tl == 1 and tr ~= 1 and td ~= 1 then --RD
            if dfr < dfd then
              g.go(right)
            else
              g.go(down)
            end
          elseif tr ~= 1 and tu ~= 1 and tl ~= 1 then --LRD
            if dfl < dfr and dfl < dfd then
              g.go(left)
            elseif dfr < dfl and dfr < dfd then
              g.go(right)
            else
              g.go(down)
            end
          end

        end
      end
    }
  end

  ghost[1].x = 13
  ghost[1].ox = 4
  ghost[1].y = 14
  ghost[1].c = {255,0,0}
  ghost[1].speed = 40
  ghost[1].idling = false
  ghost[1].inhouse = false
  ghost[1].dir = left

  ghost[2].x = 13
  ghost[2].ox = 4
  ghost[2].y = 14
  ghost[2].c = {255,170,170}
  ghost[2].speed = 40
  ghost[2].idling = false
  ghost[2].inhouse = false
  ghost[2].dir = left

  ghost[3].x = 11
  ghost[3].ox = 4
  ghost[3].y = 14
  ghost[3].c = {0,200,255}
  ghost[3].speed = 40
  ghost[3].idling = false
  ghost[3].inhouse = false
  ghost[3].dir = left

  ghost[4].x = 15
  ghost[4].ox = 4
  ghost[4].y = 14
  ghost[4].c = {255,200,0}
  ghost[4].speed = 40
  ghost[4].idling = false
  ghost[4].inhouse = false
  ghost[4].dir = left
end

function love.update(dt)
  if lv.frightcounter > 0 then
    lv.frightcounter = lv.frightcounter - dt
    if lv.frightcounter < 0 then
      lv.frightcounter = 0
      lv.mode = chase
    end
  end

  local kSpace = kb.isDown(" ")
  local kLeft = kb.isDown("left")
  local kRight = kb.isDown("right")
  local kUp = kb.isDown("up")
  local kDown = kb.isDown("down")
  local kShift = kb.isDown("lshift") or kb.isDown("rshift")

  if player.started == false then
    if kLeft then
      player.tomove = left
      player.started = true
      ghost[1].reachedTile(1)
    elseif kRight then
      player.started = true
      player.x = 14
      player.ox = -4
      player.tomove = right
      ghost[1].reachedTile(1)
    end
  else
    if kLeft and lv.map.tile[player.x-1][player.y] ~= 1 then
      if player.tomove == stationary then player.ox = 8 end
      player.tomove = left
    elseif kRight and lv.map.tile[player.x+1][player.y] ~= 1 then
      if player.tomove == stationary then player.ox = -8 end
      player.tomove = right
    elseif kUp and lv.map.tile[player.x][player.y-1] ~= 1 then
      if player.tomove == stationary then player.oy = 8 end
      player.tomove = up
    elseif kDown and lv.map.tile[player.x][player.y+1] ~= 1 then
      if player.tomove == stationary then player.oy = -8 end
      player.tomove = down
    end
  end

  if player.tomove > 0 then
    if player.ox > 0 then
      if player.ox == 8 then player.leftTile() end
      player.ox = player.ox - player.speed*dt
      if player.ox < 0 then
        player.ox = 0
        player.enteredTile()
      end
    elseif player.ox < 0 then
      if player.ox == -8 then player.leftTile() end
      player.ox = player.ox + player.speed*dt
      if player.ox > 0 then
        player.ox = 0
        player.enteredTile()
      end
    end
    if player.oy > 0 then
      if player.oy == 8 then player.leftTile() end
      player.oy = player.oy - player.speed*dt
      if player.oy < 0 then
        player.oy = 0
        player.enteredTile()
      end
    elseif player.oy < 0 then
      if player.oy == -8 then player.leftTile() end
      player.oy = player.oy + player.speed*dt
      if player.oy > 0 then
        player.oy = 0
        player.enteredTile()
      end
    end
  end

  if player.tomove > 0 then player.facing = player.tomove end

  for i, g in pairs(ghost) do
    if player.x == g.x and player.y == g.y then
      if lv.mode == frightened then
        g.eaten = true
      else
        print("DEAD!!")
      end
    end

    if g.idling == false and player.started then
      if g.ox > 0 then
        g.ox = g.ox - g.speed * dt
        if g.ox < 0 then
          g.ox = 0
          g.reachedTile(i)
        end
      elseif g.ox < 0 then
        g.ox = g.ox + g.speed * dt
        if g.ox > 0 then
          g.ox = 0
          g.reachedTile(i)
        end
      end

      if g.oy > 0 then
        g.oy = g.oy - g.speed * dt
        if g.oy < 0 then
          g.oy = 0
          g.reachedTile(i)
        end
      elseif g.oy < 0 then
        g.oy = g.oy + g.speed * dt
        if g.oy > 0 then
          g.oy = 0
          g.reachedTile(i)
        end
      end
    end
  end
end

function love.draw()
  gr.setBackgroundColor(0,0,0)

  lv.map.pellets = 0

  for x=0,28-1 do
    for y=0,36-1 do
      local t = lv.map.tile[x][y]
      if t == pellet then
        gr.setColor(255,255,255)
        gr.rectangle("fill", x*8+3,y*8+3,2,2)
        lv.map.pellets = lv.map.pellets + 1
      elseif t == power then
        if _s(time * 15) > 0 then
          gr.setColor(255,255,255)
          gr.circle("fill", x*8+4,y*8+4,4,20)
        end
        lv.map.pellets = lv.map.pellets + 1
      end
    end
  end

  gr.setColor(0,0,255)
  gr.draw(imgMap,0,0)

  local dx, dy
  if player.ox > -8 and player.ox < 8 then
    dx = player.x*8+player.ox+4
  else
    dx = player.x*8+4
  end

  if player.oy > -8 and player.oy < 8 then
    dy = player.y*8+player.oy+4
  else
    dy = player.y*8+4
  end


  for i,g in pairs(ghost) do
    gr.setColor(255,255,255)
    gr.line(g.x*8, g.y*8, g.tx*8, g.ty*8)
    if lv.mode == frightened then
      gr.setColor(0,0,255)
    else
      gr.setColor(g.c[1], g.c[2], g.c[3])
    end
    if g.eaten == false then
      local f = _f(_s(time*20)+2)
      gr.drawq(imgGhost, qGhost[f], _f(g.x*8+g.ox+4), _f(g.y*8+g.oy+4), 0, 1, 1, 8, 8)
    end
    gr.setColor(255,255,255)
    gr.drawq(imgGhost, qEyes[g.dir], _f(g.x*8+g.ox+4), _f(g.y*8+g.oy+4), 0, 1, 1, 8, 8)
  end

  local a, f
  a = _d2r(player.facing * 90 - 90)
  f = _f(_s(time*20)+2.5)

  gr.setColor(255,255,0)
  gr.drawq(imgPac, qPac[f], _f(dx), _f(dy), a, 1, 1, 8, 8)
  --gr.circle("fill",dx,dy,7,20)


  t = fps .. "\n" .. player.x .. "," .. player.y .. "\n" .. lv.map.pellets
  gr.setFont(font)
	gr.setColor(0,0,0)
	gr.print(t, 10, 1)
	gr.setColor(255,255,255)
	gr.print(t, 10, 0)
end

function love.mousepressed(x, y, button)
  if button == "l" then

  end
end

function love.keypressed(k)
  if k == " " then

  end
end

function distanceFrom(x1,y1,x2,y2) return _sq((x2 - x1) ^ 2 + (y2 - y1) ^ 2) end

function formatNumber(number)
   return (string.format("%d", number):reverse():gsub( "(%d%d%d)" , "%1," ):reverse():gsub("^(-?),","%1"))
end
